"""Asynchronous USB driver for the Crazyflie."""

from anyio import create_memory_object_stream, to_thread
from anyio.streams.stapled import StapledObjectStream
from contextlib import AsyncExitStack
from functools import partial
from typing import List, Optional

import usb

from aiocflib.utils.concurrency import Full, ThreadContext
from aiocflib.errors import NotFoundError
from aiocflib.utils.usb import (
    claim_device,
    find_devices,
    is_pyusb1,
    release_device,
    send_vendor_setup,
    USBDevice,
    USBError,
)


__author__ = "CollMot Robotics Ltd"
__all__ = ("CfUsb",)


USB_VID = 0x0483
USB_PID = 0x5740


def _find_devices() -> List[USBDevice]:
    """Returns a list of Crazyflie drones currently connected to the computer.

    This function uses `pyusb` functions directly, hence it will block the
    calling thread. You must call it in a separate worker thread when you use
    it in conjunction with an asyncio framework.
    """
    candidates = find_devices(vid=USB_VID, pid=USB_PID)
    if is_pyusb1:
        return [device for device in candidates if device.manufacturer == "Bitcraze AB"]
    else:
        return candidates


class CfUsb:
    """Low-level driver object that is used for communication with the
    Crazyflie via a USB cable.

    This object is intended to be used as an asynchronous context manager as
    follows::

        device = await CfUsb.detect_one()
        async with device as usb:
            await usb.send_bytes(b"1234")
            await usb.receive_bytes()

    In the context, a background thread is running and managing the low-level
    communication with the device; you may call `usb.send_bytes()` and
    `usb.receive_bytes()` to send raw bytes to and receive raw bytes from the
    device. The thread is terminated when the execution exits the context.

    If there is another thread that is already using the same device, the
    context blocks upon entering until the device becomes available.
    """

    _handle: Optional[USBDevice]

    @classmethod
    async def detect_all(cls):
        """Creates a list of low-level driver objects by scanning the USB buses
        for a suitable USB dongle.
        """
        return await to_thread.run_sync(_find_devices)

    @classmethod
    async def detect_one(cls, *, index: int = 0):
        """Creates a low-level driver object by scanning the USB buses for a
        suitable USB dongle, and selecting one based on the provided device
        index.

        Parameters:
            index: the index of the USB device to select if multiple dongles
                are connected

        Raises:
            IndexError: if there is no such device with the given index
        """
        devices = await cls.detect_all()
        try:
            device = devices[index]
        except IndexError:
            raise NotFoundError() from None
        return cls(device)

    def __init__(self, device: USBDevice, crtp_to_usb: bool = False):
        """Constructor.

        Creates a low-level driver object that uses the specified USB device.

        Parameters:
            device: the USB device to use
        """
        self._device = device
        self._in_queue = StapledObjectStream(*create_memory_object_stream(256))
        self._use_crtp_to_usb = bool(crtp_to_usb)

        self._receiver_thread_context = ThreadContext.create_reader(
            self._receive_bytes,
            self._in_queue,
            setup=self._configure_device,
            teardown=self._teardown_device,
        )
        self._sender_thread_context = ThreadContext.create_worker()

        self._exit_stack = None  # type: Optional[AsyncExitStack]
        self._version = None

    async def get_serial(self) -> str:
        """Returns the serial number of the associated device."""
        return await to_thread.run_sync(self._get_serial_sync)

    @property
    def use_crtp_to_usb(self):
        return self._use_crtp_to_usb

    @use_crtp_to_usb.setter
    def use_crtp_to_usb(self, value):
        self._use_crtp_to_usb = bool(value)

    @property
    def version(self) -> Optional[float]:
        """Returns the version number of the associated device.

        This property returns a valid value only after you have connected to
        the USB device.
        """
        return self._version

    async def __aenter__(self) -> "_CfUsbCommunicator":
        """Opens the driver object. This function must be called before you
        start using the driver.

        Starts an OS-level thread that will be responsible for managing
        communication over the given low-level device. Returns when the
        thread has been started.
        """
        self._version = None
        self._exit_stack = AsyncExitStack()

        stack = await self._exit_stack.__aenter__()
        await stack.enter_async_context(claim_device(self._device))
        receiver = await stack.enter_async_context(self._receiver_thread_context)
        sender = await stack.enter_async_context(self._sender_thread_context)

        sender = partial(sender, self._send_bytes)

        return _CfUsbCommunicator(sender, receiver)

    async def __aexit__(self, exc_type, exc_value, tb) -> bool:
        assert self._exit_stack is not None
        try:
            return await self._exit_stack.__aexit__(exc_type, exc_value, tb)
        finally:
            self._exit_stack = None

    def _configure_device(self):
        """Configures the USB device when the worker thread starts.

        This function is executed in the worker thread.
        """
        device = self._device

        if is_pyusb1:
            try:
                cfg = device.get_active_configuration()
            except USBError:
                cfg = None
            if cfg is None or cfg.bConfigurationValue != 1:
                device.set_configuration(1)
            handle = device
            version = float(
                "{0:x}.{1:x}".format(device.bcdDevice >> 8, device.bcdDevice & 0x0FF)
            )
        else:
            handle = device.open()
            handle.setConfiguration(1)
            handle.claimInterface(0)
            version = float(device.deviceVersion)

        self._handle, self._version = handle, version

        if self._use_crtp_to_usb:
            self._set_crtp_to_usb(True)

    def _get_serial_sync(self) -> str:
        """Retrieves the serial number of the given device in a synchronous manner."""
        # The signature for get_string has changed between versions to 1.0.0b1,
        # 1.0.0b2 and 1.0.0. Try the old signature first, if that fails try
        # the newer one.
        try:
            return usb.util.get_string(self._device, 255, self._device.iSerialNumber)
        except (USBError, ValueError):
            return usb.util.get_string(self._device, self._device.iSerialNumber)

    def _receive_bytes(self) -> Optional[bytes]:
        """Receives some data from the USB connection in a synchronous manner.

        Returns:
            the data that was received or `None` if no data was received during
            a short amount of time.

        Raises:
            IOError: when the Crazyflie was disconnected

        This function is executed in the worker thread.
        """
        assert self._handle is not None
        try:
            if is_pyusb1:
                return self._handle.read(0x81, 64, timeout=20)
            else:
                return self._handle.bulkRead(0x81, 64, 20)
        except USBError as e:
            try:
                if e.backend_error_code == -7 or e.backend_error_code == -116:
                    # Normal, the read was empty
                    pass
                else:
                    raise IOError("Crazyflie disconnected")
            except AttributeError:
                # pyusb < 1.0 doesn't implement getting the underlying error
                # number and it seems as if it's not possible to detect
                # if the cable is disconnected. So this detection is not
                # supported, but the "normal" case will work.
                pass

        return None

    def _send_bytes(self, data: bytes) -> None:
        """Sends some data via the USB connection in a synchronous manner.

        This function is executed in the worker thread.

        Raises:
            IOError: when the Crazyflie was disconnected
        """
        assert self._handle is not None
        try:
            if is_pyusb1:
                self._handle.write(endpoint=1, data=data, timeout=20)
            else:
                self._handle.bulkWrite(1, data, 20)
        except USBError as ex:
            raise IOError("Error while sending packet") from ex

    def _teardown_device(self, exc_type, exc_value, tb):
        """Tears down the connection to the USB device when the worker thread
        exits.

        This function is executed in the worker thread.
        """
        try:
            if self._use_crtp_to_usb:
                self._set_crtp_to_usb(False)
        except Exception:
            # maybe the device was disconnected already?
            pass

        try:
            release_device(self._handle)
        finally:
            self._handle = None
            self._version = None

    def _set_crtp_to_usb(self, value):
        send_vendor_setup(self._handle, 0x01, 0x01, int(bool(value)))


class _CfUsbCommunicator:
    """Object that is returned when entering a CfUsb context and that allows
    us to send packets to and receive packets from the USB connection.

    This is an internal class; you do not need to construct it yourself.
    """

    def __init__(self, sender, receiver):
        """Constructor.

        Parameters:
            sender: an async function that can be used to send a request to the
                outbound worker thread to send some data and wait for its result
            receiver: an async function that can be used to send a request to the
                inbound worker thread to receive the next packet
        """
        self._sender = sender
        self._receiver = receiver

    async def send_bytes(self, data: bytes) -> None:
        """Sends some bytes to the connected Crazyflie via the USB port.

        Parameters:
            data: the data to send to the Crazyflie

        Raises:
            IOError: if the request queue to the USB outbound thread is full or
                when the Crazyflie was disconnected
        """
        try:
            return await self._sender(data)
        except Full:
            raise IOError("Request queue to USB outbound thread is full") from None

    async def receive_bytes(self) -> bytes:
        """Receives some bytes from the connected Crazyflie via the USB port.

        Returns:
            the data that was received. It is guaranteed to have at least one
            byte in it.

        Raises:
            IOError: when the Crazyflie was disconnected
        """
        while True:
            data = await self._receiver()
            if data:
                return data


async def test():
    device = await CfUsb.detect_one()
    device.use_crtp_to_usb = True
    async with device as usb:
        # \xfd\x01 sends a "get version" command to the link control port
        await usb.send_bytes(b"\xfd\x01")
        data = await usb.receive_bytes()
        print("Received:", data)


if __name__ == "__main__":
    import trio

    trio.run(test)



"""Classes related to handling platform service messages of a Crazyflie."""

from typing import AsyncIterator

from aiocflib.crtp.crtpstack import CRTPPacket, CRTPPort

from .crazyflie import Crazyflie
from .platform import PlatformChannel

__all__ = ("AppChannel",)


class AppChannel:
    """Class representing the handler of app channel messages for a Crazyflie
    instance.
    """

    _crazyflie: Crazyflie

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle the app channel
                messages
        """
        self._crazyflie = crazyflie

    async def packets(self) -> AsyncIterator[CRTPPacket]:
        """Async generator that yields messages coming on the app-specific
        channel from a Crazyflie.
        """
        async for packet in self._crazyflie.packets(port=CRTPPort.PLATFORM):
            if packet.channel == PlatformChannel.APP_CHANNEL:
                yield packet



"""Functions for broadcasting packets that are typically sent to multiple
Crazyflie drones using a single radio.
"""

from typing import Sequence, Tuple, TYPE_CHECKING

from aiocflib.crtp import CRTPPort
from aiocflib.utils.quaternion import QuaternionXYZW

from .localization import GenericLocalizationCommand, Localization, LocalizationChannel

if TYPE_CHECKING:
    from aiocflib.crtp.broadcaster import _Broadcaster


async def broadcast_external_position_packed(
    broadcaster: "_Broadcaster", items: Sequence[Tuple[int, Tuple[float, float, float]]]
) -> None:
    """Broadcasts a packet containing external position information for multiple
    Crazyflies using the given broadcaster.

    Parameters:
        broadcaster: the broadcaster to use
        items: a sequence of pairs containing a numeric Crazyflie ID
            (the last byte of its radio address) and a 3D coordinate.
            Coordinates must be less than ~32.7 meters in absolute value.
            At most four items fit into a single Crazyflie CRTP packet.
    """
    data = Localization.encode_external_position_packed(items)
    await broadcaster.send_packet(
        port=CRTPPort.LOCALIZATION,
        channel=LocalizationChannel.POSITION_PACKED,
        data=data,
    )


async def broadcast_external_pose_packed(
    broadcaster: "_Broadcaster",
    items: Sequence[Tuple[int, Tuple[float, float, float], QuaternionXYZW]],
) -> None:
    """Broadcasts a packet containing external pose (position + attitude)
    information for multiple Crazyflies using the given broadcaster.

    Parameters:
        broadcaster: the broadcaster to use
        items: a sequence of triplets containing a numeric Crazyflie ID
            (the last byte of its radio address), a 3D coordinate and a
            4D quaternion in XYZW order. Coordinates must be less than ~32.7
            meters in absolute value. At most two items fit into a single
            Crazyflie CRTP packet.
    """
    data = Localization.encode_external_pose_packed(items)
    await broadcaster.send_packet(
        port=CRTPPort.LOCALIZATION,
        channel=LocalizationChannel.GENERIC,
        data=bytes([GenericLocalizationCommand.EXT_POSE_PACKED]) + data,
    )


async def broadcast_emergency_stop(broadcaster: "_Broadcaster") -> None:
    """Broadcasts an emergency stop packet for multiple Crazyflies using the
    given broadcaster.

    Parameters:
        broadcaster: the broadcaster to use
    """
    await broadcaster.send_packet(
        port=CRTPPort.LOCALIZATION,
        channel=LocalizationChannel.GENERIC,
        data=bytes([GenericLocalizationCommand.ENABLE_EMERGENCY_STOP]),
    )



"""Checksum-related utility functions."""

from binascii import crc32 as crc32_uint32
from struct import Struct

__all__ = ("crc32", "ensure_not_all_zero_bytes")


_uint32_struct = Struct("<I")


def crc32(data: bytes) -> bytes:
    """Calculates the CRC32 checksum of the given bytes object.

    Parameters:
        data: the data to calculate the checksum for

    Returns:
        the CRC32 checksum, in little endian order
    """
    return _uint32_struct.pack(crc32_uint32(data) & 0xFFFFFFFF)


_sentinel = b"\x0b\xad\xca\xfe" * 64


def ensure_not_all_zero_bytes(data: bytes) -> bytes:
    """Examines the given bytes object and replaces it with a sentinel value
    if it has at least one byte and all the bytes in the object are zeros.

    This is used internally by MemoryHandler.write_with_checksum(); you probably
    don't need it on your own.

    Parameters:
        data: the input data

    Returns:
        the input data if it contains at least one non-zero byte or if it is
        empty, otherwise a sentinel value
    """
    if data and not any(x for x in data):
        if len(data) <= len(_sentinel):
            return _sentinel[: len(data)]
        else:
            result = _sentinel * (len(data) // len(_sentinel))
            result += _sentinel[: len(data) - len(result)]
            return result
    else:
        return data



"""Middleware-related utility functions."""

from ..drivers.base import CRTPDriver

__all__ = ("unwrap_middleware",)


def unwrap_middleware(driver: CRTPDriver) -> CRTPDriver:
    """Takes a CRTP driver possibly wrapped in CRTP middleware, and returns
    the innermost driver without all the wrapping middleware.
    """
    while hasattr(driver, "wrapped"):
        driver = driver.wrapped
    return driver



from dataclasses import dataclass, field
from enum import IntEnum
from errno import EIO
from struct import Struct
from typing import (
    Any,
    ClassVar,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

from aiocflib.crtp.crtpstack import MemoryType

from .crazyflie import Crazyflie
from .mem import MemoryHandler

__all__ = ("LighthouseBsCalibration", "LighthouseBsGeometry", "LighthouseConfiguration")


Vector3D = Tuple[float, float, float]
Matrix3D = Tuple[Vector3D, Vector3D, Vector3D]


@dataclass(frozen=True)
class LighthouseBsGeometry:
    """Container for geometry data of one Lighthouse base station"""

    origin: Vector3D = (0, 0, 0)
    rotation_matrix: Matrix3D = ((0, 0, 0), (0, 0, 0), (0, 0, 0))
    valid: bool = False

    _struct: ClassVar[Struct] = Struct("<ffffffffffff?")
    size_in_bytes: ClassVar[int] = _struct.size

    @classmethod
    def from_bytes(cls, data: bytes):
        """Constructs a Lighthouse base station geometry object from its raw
        byte-level representation from the Lighthouse memory of a Crazyflie.

        Parameters:
            data: the data to unpack from

        Returns:
            the unpacked object
        """
        if len(data) != cls.size_in_bytes:
            raise ValueError(
                f"invalid length, expected {cls.size_in_bytes} bytes, got {len(data)}"
            )

        result, _ = cls.unpack_from_bytes(data)
        return result

    @classmethod
    def from_json(cls, obj: Dict[str, Any]):
        """Constructs a Lighthouse base station geometry object from its JSON
        object representation created earlier with `to_json()`.
        """
        assert len(obj["origin"]) == 3
        assert len(obj["rotation"]) == 3 and all(
            len(row) == 3 for row in obj["rotation"]
        )
        return cls(
            origin=tuple(obj["origin"]),  # type: ignore
            rotation_matrix=tuple(tuple(row) for row in obj["rotation"]),  # type: ignore
            valid=True,
        )

    @classmethod
    def unpack_from_bytes(
        cls: Type["LighthouseBsGeometry"], data: bytes, offset: int = 0
    ) -> Tuple["LighthouseBsGeometry", int]:
        """Constructs a Lighthouse sweep calibration object from its raw
        byte-level representation from the Lighthouse memory of a Crazyflie.

        Parameters:
            data: the data to unpack from
            offset: optional offset into the data object

        Returns:
            the unpacked object and the index of the first _unconsumed_ byte
            from the incoming data
        """
        if len(data) - offset < cls.size_in_bytes:
            raise ValueError(
                f"invalid length, expected {cls.size_in_bytes} bytes, got {len(data) - offset}"
            )

        items = cls._struct.unpack_from(data, offset)
        origin = cast(Vector3D, items[:3])
        rotation_matrix = cast(Matrix3D, (items[3:6], items[6:9], items[9:12]))
        valid = items[12]

        return (
            cls(origin=origin, rotation_matrix=rotation_matrix, valid=valid),
            offset + cls.size_in_bytes,
        )

    def to_bytes(self) -> bytes:
        """Converts the Lighthouse base station geometry object into a raw
        byte-level representation used in the Lighthouse memory.
        """
        items: List[Any] = []
        items.extend(self.origin)
        items.extend(sum(self.rotation_matrix, ()))
        items.append(self.valid)
        return self._struct.pack(*items)

    def to_json(self) -> Dict[str, Any]:
        """Converts the Lighthouse base station data into a Python object that
        can be written directly into a JSON or YAML file.
        """
        return {"origin": self.origin, "rotation": self.rotation_matrix}


@dataclass(frozen=True)
class LighthouseCalibrationSweep:
    """Container for calibration data of a single sweep plane of a Lighthouse
    base station.
    """

    phase: float = 0.0
    tilt: float = 0.0
    curve: float = 0.0
    gibmag: float = 0.0
    gibphase: float = 0.0
    ogeemag: float = 0.0
    ogeephase: float = 0.0

    _struct: ClassVar[Struct] = Struct("<fffffff")
    size_in_bytes: ClassVar[int] = _struct.size

    @classmethod
    def from_bytes(cls, data: bytes):
        """Constructs a Lighthouse sweep calibration object from its raw
        byte-level representation from the Lighthouse memory of a Crazyflie.

        Parameters:
            data: the data to unpack from

        Returns:
            the unpacked object
        """
        if len(data) != cls.size_in_bytes:
            raise ValueError(
                f"invalid length, expected {cls.size_in_bytes} bytes, got {len(data)}"
            )

        obj, _ = cls.unpack_from_bytes(data)
        return obj

    @classmethod
    def from_json(cls, obj: Dict[str, Any]):
        """Constructs a Lighthouse sweep calibration object from its JSON
        object representation created earlier with `to_json()`.
        """
        return cls(
            phase=obj["phase"],
            tilt=obj["tilt"],
            curve=obj["curve"],
            gibmag=obj["gibmag"],
            gibphase=obj["gibphase"],
            ogeemag=obj["ogeemag"],
            ogeephase=obj["ogeephase"],
        )

    @classmethod
    def unpack_from_bytes(cls, data: bytes, offset: int = 0):
        """Constructs a Lighthouse sweep calibration object from its raw
        byte-level representation from the Lighthouse memory of a Crazyflie.

        Parameters:
            data: the data to unpack from
            offset: optional offset into the data object

        Returns:
            the unpacked object and the index of the first _unconsumed_ byte
            from the incoming data
        """
        if len(data) - offset < cls.size_in_bytes:
            raise ValueError(
                f"invalid length, expected {cls.size_in_bytes} bytes, got {len(data) - offset}"
            )

        items = cls._struct.unpack_from(data, offset)
        return (
            cls(
                phase=items[0],
                tilt=items[1],
                curve=items[2],
                gibmag=items[3],
                gibphase=items[4],
                ogeemag=items[5],
                ogeephase=items[6],
            ),
            offset + cls.size_in_bytes,
        )

    def to_bytes(self) -> bytes:
        """Converts the Lighthouse sweep calibration object into a raw byte-level
        representation used in the Lighthouse memory.
        """
        return self._struct.pack(
            self.phase,
            self.tilt,
            self.curve,
            self.gibmag,
            self.gibphase,
            self.ogeemag,
            self.ogeephase,
        )

    def to_json(self) -> Dict[str, Any]:
        """Converts the Lighthouse sweep calibration data into a Python object that
        can be written directly into a JSON or YAML file.
        """
        return {
            "phase": self.phase,
            "tilt": self.tilt,
            "curve": self.curve,
            "gibmag": self.gibmag,
            "gibphase": self.gibphase,
            "ogeemag": self.ogeemag,
            "ogeephase": self.ogeephase,
        }


@dataclass(frozen=True)
class LighthouseBsCalibration:
    """Container for calibration data of one Lighthouse base station."""

    sweeps: Tuple[LighthouseCalibrationSweep, LighthouseCalibrationSweep] = field(
        default=(LighthouseCalibrationSweep(), LighthouseCalibrationSweep())
    )
    uid: int = 0
    valid: bool = False

    _struct: ClassVar[Struct] = Struct("<L?")
    size_in_bytes: ClassVar[int] = (
        _struct.size + 2 * LighthouseCalibrationSweep.size_in_bytes
    )

    @classmethod
    def from_bytes(cls, data: bytes):
        """Constructs a Lighthouse base station calibration object from its raw
        byte-level representation from the Lighthouse memory of a Crazyflie.

        Parameters:
            data: the data to unpack from

        Returns:
            the unpacked object
        """
        if len(data) != cls.size_in_bytes:
            raise ValueError(
                f"invalid length, expected {cls.size_in_bytes} bytes, got {len(data)}"
            )

        obj, _ = cls.unpack_from_bytes(data)
        return obj

    @classmethod
    def from_json(cls, obj: Dict[str, Any]):
        """Constructs a Lighthouse base station calibration object from its JSON
        object representation created earlier with `to_json()`.
        """
        assert len(obj["sweeps"]) == 2
        return cls(
            uid=int(obj["uid"]),
            sweeps=tuple(  # type: ignore
                LighthouseCalibrationSweep.from_json(item) for item in obj["sweeps"]
            ),
            valid=True,
        )

    @classmethod
    def unpack_from_bytes(cls, data: bytes, offset: int = 0):
        """Constructs a Lighthouse base station calibration object from its raw
        byte-level representation from the Lighthouse memory of a Crazyflie.

        Parameters:
            data: the data to unpack from
            offset: optional offset into the data object

        Returns:
            the unpacked object and the index of the first _unconsumed_ byte
            from the incoming data
        """
        if len(data) - offset < cls.size_in_bytes:
            raise ValueError(
                f"invalid length, expected {cls.size_in_bytes} bytes, got {len(data) - offset}"
            )

        sweep1, offset = LighthouseCalibrationSweep.unpack_from_bytes(data, offset)
        sweep2, offset = LighthouseCalibrationSweep.unpack_from_bytes(data, offset)

        items = cls._struct.unpack_from(data, offset)
        return (
            cls(sweeps=(sweep1, sweep2), uid=items[0], valid=items[1]),
            offset + cls._struct.size,
        )

    def to_bytes(self) -> bytes:
        """Converts the Lighthouse base station calibration object into a raw
        byte-level representation used in the Lighthouse memory.
        """
        return b"".join(
            [
                self.sweeps[0].to_bytes(),
                self.sweeps[1].to_bytes(),
                self._struct.pack(self.uid, self.valid),
            ]
        )

    def to_json(self) -> Dict[str, Any]:
        """Converts the Lighthouse base station calibration data into a Python
        object that can be written directly into a JSON or YAML file.
        """
        return {
            "sweeps": tuple(sweep.to_json() for sweep in self.sweeps),
            "uid": self.uid,
        }


class LighthouseSystemType(IntEnum):
    """Enum representing the different Lighthouse system versions that we
    support.
    """

    UNKNOWN = 0
    V1 = 1
    V2 = 2

    def describe(self) -> str:
        """Returns a human-readable description of the system type."""
        if int(self) in (1, 2):
            return f"Lighthouse v{int(self)}"
        else:
            return "unknown'"


@dataclass(frozen=True)
class LighthouseConfiguration:
    """Data class that encapsulates the geometry _and_ calibration data of
    all base stations in a Lighthouse system.
    """

    system_type: LighthouseSystemType = LighthouseSystemType.UNKNOWN
    calibrations: Dict[int, LighthouseBsCalibration] = field(default_factory=dict)
    geometries: Dict[int, LighthouseBsGeometry] = field(default_factory=dict)

    @property
    def bs_ids(self) -> FrozenSet[int]:
        return frozenset(self.calibrations.keys()) | frozenset(self.geometries.keys())

    @property
    def valid_bs_ids(self) -> FrozenSet[int]:
        return frozenset(
            k for k, v in self.calibrations.items() if v.valid
        ) & frozenset(k for k, v in self.geometries.items() if v.valid)

    def to_json(self) -> Dict[str, Any]:
        """Converts the Lighthouse configuration into a Python object that can
        be written directly into a JSON or YAML file.
        """
        result = {
            "version": "1",
            "type": "lighthouse_system_configuration",
            "calibs": {k: v.to_json() for k, v in self.calibrations.items()},
            "geos": {k: v.to_json() for k, v in self.geometries.items()},
        }
        if self.system_type != LighthouseSystemType.UNKNOWN:
            result["systemType"] = int(self.system_type)
        return result

    @property
    def valid(self) -> bool:
        """Returns whether the Lighthouse configuration is valid.

        A configuration is valid if it has a valid system type, it has at least
        one base station and all the calibrations and geometries are valid.
        """
        return (
            self.system_type != LighthouseSystemType.UNKNOWN
            and all(calib.valid for calib in self.calibrations.values())
            and all(geo.valid for geo in self.geometries.values())
            and len(set(self.calibrations.keys()) & set(self.geometries.keys())) > 0
        )


class Lighthouse:
    """Class representing the Lighthouse subsystem of a Crazyflie instance."""

    CALIB_START_ADDR: ClassVar[int] = 0x1000
    GEO_START_ADDR: ClassVar[int] = 0
    PAGE_SIZE: ClassVar[int] = 0x100

    _crazyflie: Crazyflie
    _mem: Optional[MemoryHandler]

    number_of_base_stations: int

    def __init__(self, crazyflie: Crazyflie, *, number_of_base_stations: int = 2):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie instance
        """
        self._crazyflie = crazyflie
        self._mem = None
        self.number_of_base_stations = number_of_base_stations

    async def clear_all(self) -> None:
        """Clears the calibration and geometry data of all base stations on
        the Crazyflie.
        """
        await self.clear_calibrations()
        await self.clear_geometries()

    async def clear_calibration(self, index: int) -> None:
        """Clears the calibration data of a single base station on the Crazyflie."""
        await self.set_calibration(index, LighthouseBsCalibration())

    async def clear_calibrations(self, indices: Optional[Iterable[int]] = None) -> None:
        """Clears the calibration data of multiple base stations on the Crazyflie.

        Parameters:
            indices: the indices of the base stations to clear; `None` means to
                clear all base stations.
        """
        if indices is None:
            ignore_errors = True
            indices = range(self.number_of_base_stations)
        else:
            ignore_errors = False

        for index in indices:
            try:
                await self.clear_calibration(index)
            except IOError as ex:
                if ex.errno == EIO and ignore_errors:
                    # Probably just an invalid base station ID
                    continue
                else:
                    raise

    async def clear_geometry(self, index: int) -> None:
        """Clears the geometry data of a single base station on the Crazyflie."""
        await self.set_geometry(index, LighthouseBsGeometry())

    async def clear_geometries(
        self, indices: Optional[Iterable[int]] = None, *, ignore_errors: bool = False
    ) -> None:
        """Clears the geometry data of multiple base stations on the Crazyflie.

        Parameters:
            indices: the indices of the base stations to clear; `None` means to
                clear all base stations.
        """
        if indices is None:
            ignore_errors = True
            indices = range(self.number_of_base_stations)
        else:
            ignore_errors = False

        for index in indices:
            try:
                await self.clear_geometry(index)
            except IOError as ex:
                if ex.errno == EIO and ignore_errors:
                    # Probably just an invalid base station ID
                    continue
                else:
                    raise

    async def get_calibration(self, index: int) -> Optional[LighthouseBsCalibration]:
        """Retrieves the calibration data of a single base station from the
        Crazyflie.

        Parameters:
            index: the index of the base station

        Returns: the calibration data of a single base station from the Crazyflie,
            or `None` if the calibration data is not valid for the given base
            station index.
        """
        mem = await self._get_memory()
        try:
            data = await mem.read(
                self._get_address_of_bs_calibration(index),
                LighthouseBsCalibration.size_in_bytes,
            )
        except IOError as ex:
            if ex.errno == EIO:
                # This is okay, probably the base station index is invalid
                return None
            else:
                raise

        calibration = LighthouseBsCalibration.from_bytes(data)
        return calibration if calibration.valid else None

    async def get_calibrations(self) -> Dict[int, LighthouseBsCalibration]:
        """Returns the calibration data of all the base stations from the Crazyflie."""
        result: Dict[int, LighthouseBsCalibration] = {}
        for i in range(self.number_of_base_stations):
            calibration = await self.get_calibration(i)
            if calibration:
                result[i] = calibration
        return result

    async def get_configuration(self) -> LighthouseConfiguration:
        """Returns the Lighthouse configuration of the Crazyflie."""
        calib = await self.get_calibrations()
        geos = await self.get_geometries()
        system_type = await self.get_system_type()
        return LighthouseConfiguration(
            system_type=system_type, calibrations=calib, geometries=geos
        )

    async def get_geometry(self, index: int) -> Optional[LighthouseBsGeometry]:
        """Retrieves the geometry of a single base station from the Crazyflie.

        Parameters:
            index: the index of the base station

        Returns: the geometry of the base station or `None` if the geometry is
            not valid for the given base station index.
        """
        mem = await self._get_memory()
        try:
            data = await mem.read(
                self._get_address_of_bs_geometry(index),
                LighthouseBsGeometry.size_in_bytes,
            )
        except IOError as ex:
            if ex.errno == EIO:
                # This is okay, probably the base station index is invalid
                return None
            else:
                raise

        geometry = LighthouseBsGeometry.from_bytes(data)
        return geometry if geometry.valid else None

    async def get_geometries(self) -> Dict[int, LighthouseBsGeometry]:
        """Returns the geometries of all the base stations from the Crazyflie."""
        result: Dict[int, LighthouseBsGeometry] = {}
        for i in range(self.number_of_base_stations):
            geometry = await self.get_geometry(i)
            if geometry:
                result[i] = geometry
        return result

    async def get_system_type(self) -> LighthouseSystemType:
        """Returns the configured Lighthouse system type of the Crazyflie."""
        value = await self._crazyflie.param.get("lighthouse.systemType")
        try:
            return LighthouseSystemType(int(value))
        except Exception:
            return LighthouseSystemType.UNKNOWN

    async def persist(self) -> None:
        """Copies the current calibration and geometry data on the Crazyflie to
        persistent storage so it gets loaded the next time the Crazyflie boots.
        """
        await self._crazyflie.localization.persist_lighthouse_data()

    async def set_calibration(
        self,
        index: int,
        calibration: LighthouseBsCalibration,
    ) -> None:
        """Sets the calibration data of the base station with the given index.

        Note that this function does _not_ persist the calibration data on the
        Crazyflie into permanent memory. Call `persist()` to make sure that the
        changes are permanent.

        Parameters:
            index: the index of the base station
            calibration: the calibration data to set on the Crazyflie
        """
        mem = await self._get_memory()
        await mem.write(
            self._get_address_of_bs_calibration(index), calibration.to_bytes()
        )

    async def set_calibrations(self, data: Dict[int, LighthouseBsCalibration]) -> None:
        """Sets the calibration data of multiple base stations.

        Parameters:
            data: a dictionary mapping base station IDs to calibration data
        """
        for index, calibration in data.items():
            await self.set_calibration(index, calibration)

    async def set_configuration(self, config: LighthouseConfiguration) -> None:
        """Replaces the entire Lighthouse base station configuration with the
        given object.

        Note that this function does _not_ persist the geometry data on the
        Crazyflie into permanent memory. Call `persist()` to make sure that the
        changes are permanent.
        """
        if not config.valid:
            raise RuntimeError("Configuration is invalid")

        await self.clear_all()
        await self._crazyflie.param.set(
            "lighthouse.systemType", int(config.system_type)
        )
        for bs_id in config.valid_bs_ids:
            await self.set_calibration(bs_id, config.calibrations[bs_id])
            await self.set_geometry(bs_id, config.geometries[bs_id])
        await self._crazyflie.param.set("lighthouse.bsCalibReset", 1)
        await self.persist()

    async def set_geometry(
        self,
        index: int,
        geometry: LighthouseBsGeometry,
    ) -> None:
        """Sets the geometry object of the base station with the given index.

        Note that this function does _not_ persist the geometry data on the
        Crazyflie into permanent memory. Call `persist()` to make sure that the
        changes are permanent.

        Parameters:
            index: the index of the base station
            geometry: the geometry data to set on the Crazyflie
        """
        mem = await self._get_memory()
        await mem.write(self._get_address_of_bs_geometry(index), geometry.to_bytes())

    async def set_geometries(self, data: Dict[int, LighthouseBsGeometry]) -> None:
        """Sets the geometry data of multiple base stations.

        Parameters:
            data: a dictionary mapping base station IDs to calibration data
        """
        for index, geometry in data.items():
            await self.set_geometry(index, geometry)

    def _get_address_of_bs_calibration(self, index: int) -> int:
        """Returns the address of the calibration data of the base station with
        the given index.
        """
        return self.CALIB_START_ADDR + index * self.PAGE_SIZE

    def _get_address_of_bs_geometry(self, index: int) -> int:
        """Returns the address of the geometry data of the base station with
        the given index.
        """
        return self.GEO_START_ADDR + index * self.PAGE_SIZE

    async def _get_memory(self) -> MemoryHandler:
        """Returns the memory handler object of the Lighthouse memory."""
        if self._mem is None:
            self._mem = await self._crazyflie.mem.find(MemoryType.LIGHTHOUSE)
        return self._mem


async def test():
    from pprint import pprint

    uri = "radio+log://0/80/2M/E7E7E7E701"
    async with Crazyflie(uri) as cf:
        cf.lighthouse.number_of_base_stations = 16

        calibration = await cf.lighthouse.get_calibrations()
        print("Calibration:")
        pprint(calibration)
        print("")

        geom = await cf.lighthouse.get_geometries()
        print("Geometry:")
        pprint(geom)

        await cf.lighthouse.clear_all()

        calibration = await cf.lighthouse.get_calibrations()
        print("Calibration:")
        pprint(calibration)
        print("")

        geom = await cf.lighthouse.get_geometries()
        print("Geometry:")
        pprint(geom)

        await cf.lighthouse.persist()


if __name__ == "__main__":
    from aiocflib.crtp import init_drivers
    import trio

    init_drivers()
    try:
        trio.run(test)
    except KeyboardInterrupt:
        pass



"""USB-related low-level utility functions."""

from anyio import Lock
from contextlib import asynccontextmanager
from typing import Any, List
from weakref import WeakValueDictionary

import os

__all__ = (
    "claim_device",
    "find_devices",
    "get_vendor_setup",
    "is_pyusb1",
    "release_device",
    "send_vendor_setup",
    "USBDevice",
    "USBError",
)

#: Type variable to represent a USB device
USBDevice = Any

#: Module-level variable to hold a mapping from unique USB device IDs to their
#: locks that prevent concurrent access to these devices
_locks: "WeakValueDictionary[str, Lock]" = WeakValueDictionary()

try:
    import usb.core

    _backend = None
    if os.name == "nt":
        import usb.backend.libusb0 as libusb0

        _backend = libusb0.get_backend()
    is_pyusb1 = True

except Exception:
    _backend = None
    is_pyusb1 = False


@asynccontextmanager
async def claim_device(device: USBDevice):
    """Asynchronous context manager that claims a USB device and prevents
    others from claiming the same device as long as the execution stays
    within the context.
    """
    global _locks

    uid = get_device_uid(device)
    lock = _locks.get(uid)
    if lock is None:
        _locks[uid] = lock = Lock()

    async with lock:
        yield


def find_devices(vid: int, pid: int) -> List[USBDevice]:
    """Helper function that finds all USB devices with a given vendor and
    product ID on all available USB buses.
    """
    if is_pyusb1:
        return list(
            usb.core.find(idVendor=vid, idProduct=pid, find_all=1, backend=_backend)  # type: ignore
        )
    else:
        return [
            device
            for bus in usb.busses()  # type: ignore
            for device in bus.devices
            if device.idVendor == vid and device.idProduct == pid
        ]


def get_vendor_setup(handle, request, value, index, length, timeout=1000):
    if is_pyusb1:
        return handle.ctrl_transfer(
            usb.TYPE_VENDOR | 0x80,  # type: ignore
            request,
            wValue=value,
            wIndex=index,
            timeout=timeout,
            data_or_wLength=length,
        )
    else:
        return handle.controlMsg(
            usb.TYPE_VENDOR | 0x80,  # type: ignore
            request,
            length,
            value=value,
            index=index,
            timeout=timeout,
        )


def get_device_uid(device: USBDevice) -> str:
    """Returns a string that can be used to uniquely identify a USB device
    that is currently plugged in.

    Currently the string contains the bus number and the device number of the
    device. This means that the unique identifier may change if the device is
    unplugged and then plugged in again at a different port.
    """
    return "{0.bus}:{0.address}".format(device)


def release_device(device: USBDevice):
    from usb.util import dispose_resources

    dispose_resources(device)


def send_vendor_setup(handle, request, value, index=0, data=(), timeout=1000):
    if is_pyusb1:
        handle.ctrl_transfer(
            usb.TYPE_VENDOR,  # type: ignore
            request,
            wValue=value,
            wIndex=index,
            timeout=timeout,
            data_or_wLength=data,
        )
    else:
        handle.controlMsg(
            usb.TYPE_VENDOR,  # type: ignore
            request,
            data,
            value=value,
            index=index,
            timeout=timeout,
        )


USBError = usb.core.USBError  # type: ignore



"""Asynchronous USB driver for the Crazyflie."""

from anyio import connect_tcp
from anyio.abc import ByteStream
from anyio.streams.buffered import BufferedByteReceiveStream
from array import array
from contextlib import AsyncExitStack
from functools import partial
from typing import Optional


__author__ = "CollMot Robotics Ltd"
__all__ = ("SITL",)


class SITL:
    r"""Low-level driver object that is used for communication with the
    Crazyflie software-in-the-loop simulator via a TCP connection.

    This object is intended to be used as an asynchronous context manager as
    follows::

        async with SITL(port=5432) as sitl:
            await sitl.send_bytes(b"\xfd\x01")
            await sitl.receive_bytes()
    """

    def __init__(self, host: str = "localhost", port: int = 5432):
        """Constructor.

        Creates a low-level driver object that communicates with a Crazyflie
        SITL simulator at the specified hostname and port.

        Parameters:
            host: the hostname to connect to
            port: the port to connect to
        """
        self._address = host, port
        self._exit_stack = None  # type: Optional[AsyncExitStack]
        self._client = None

    async def __aenter__(self):
        """Opens the driver object. This function must be called before you
        start using the driver.
        """
        self._exit_stack = AsyncExitStack()

        await self._exit_stack.__aenter__()
        tcp_context = await connect_tcp(*self._address)
        client = await self._exit_stack.enter_async_context(tcp_context)
        client = BufferedByteReceiveStream(client)

        return _SITLCommunicator(
            partial(self._send_bytes, client), partial(self._receive_bytes, client)
        )

    async def __aexit__(self, exc_type, exc_value, tb) -> bool:
        assert self._exit_stack is not None
        try:
            return await self._exit_stack.__aexit__(exc_type, exc_value, tb)
        finally:
            self._exit_stack = None

    async def _receive_bytes(
        self, client: BufferedByteReceiveStream
    ) -> Optional[array]:
        """Receives some data from the SITL connection in a synchronous manner.

        Parameters:
            client: the TCP client to read data from

        Returns:
            the data that was received

        Raises:
            IOError: when the simulator was disconnected
        """
        length = ord(await client.receive_exactly(1))
        return await client.receive_exactly(length)

    async def _send_bytes(self, client: ByteStream, data: bytes) -> None:
        """Sends some data via the TCP connection in a synchronous manner.

        Parameters:
            client: the TCP client to send data to
            data: the data to send

        Raises:
            IOError: when the simulator was disconnected
        """
        length = len(data)
        assert length <= 64
        await client.send_all(bytes((length,)) + data)


class _SITLCommunicator:
    """Object that is returned when entering a SITL context and that allows
    us to send packets to and receive packets from the SITL connection.

    This is an internal class; you do not need to construct it yourself.
    """

    def __init__(self, sender, receiver):
        """Constructor.

        Parameters:
            sender: an async function that can be used to send some data to the
                simulator
            receiver: an async function that can be used to receive the next
                packet from the simulator
        """
        self._sender = sender
        self._receiver = receiver

    async def send_bytes(self, data: array) -> None:
        """Sends some bytes to the connected simulator via the TCP connection.

        Parameters:
            data: the data to send to the simulator

        Raises:
            IOError: if the simulator was shut down or there was some other IO
                error during sending
        """
        return await self._sender(data)

    async def receive_bytes(self) -> array:
        """Receives some bytes from the connected simulator via the TCP
        connection.

        Returns:
            the data that was received. It is guaranteed to have at least one
            byte in it.

        Raises:
            IOError: if the simulator was shut down or there was some other IO
                error during receiving
        """
        while True:
            data = await self._receiver()
            if data:
                return data


async def test():
    async with SITL() as sitl:
        # \xfd\x01 sends a "get version" command to the link control port
        await sitl.send_bytes(b"\xfd\x01")
        data = await sitl.receive_bytes()
        print("Received:", data)


if __name__ == "__main__":
    import trio

    trio.run(test)



from typing import Callable

from aiocflib.crtp.drivers.base import CRTPDriver
from aiocflib.utils.registry import Registry

__all__ = ("MiddlewareRegistry", "find", "register")

#: Type specification for CRTP driver middleware
Middleware = Callable[[CRTPDriver], CRTPDriver]

#: Mapping that maps names to the corresponding middleware classes
MiddlewareRegistry = Registry()  # type: Registry[Middleware]

find = MiddlewareRegistry.find
"""Returns the registered middleware corresponding to the given name.

Parameters:
    name: the name to which the middleware is registered

Raises:
    KeyError: if there is no registered middleware for the given name

Returns:
    the middleware corresponding to the name
"""

register = MiddlewareRegistry.register
"""Class decorator factory that returns a decorator that registers a class
as a middleware with the given name.

Parameters:
    name: the name for which the driver will be registered

Returns:
    an appropriate decorator that can then be applied to a middleware class
"""



from anyio import create_memory_object_stream, move_on_after, WouldBlock
from contextlib import asynccontextmanager
from urllib.parse import urlparse

from aiocflib.crtp.crtpstack import CRTPPacket
from aiocflib.drivers.sitl import SITL
from aiocflib.utils.concurrency import create_daemon_task_group, ObservableValue

from .base import CRTPDriver
from .registry import register

from ..strategies import (
    BackoffPollingStrategy,
    DefaultPollingStrategy,
    NoPollingStrategy,
)

_link_quality = None  # type: Optional[ObservableValue[float]]


@register("sitl")
class SITLDriver(CRTPDriver):
    """CRTP driver that allows us to communicate with a simulated Crazyflie that
    is running in a software-in-the-loop simulator and is accessible via a
    TCP connection.

    Attributes:
        polling_strategy: a callable that decides how often we should poll the
            downlink if there are no packets that we want to send to the
            Crazyflie
    """

    PRESETS = {
        "default": (DefaultPollingStrategy,),
        "patient": (BackoffPollingStrategy,),
        "noPolling": (NoPollingStrategy,),
    }

    @asynccontextmanager
    async def _connected_to(self, uri: str):
        parts = urlparse(uri)
        host, _, port = parts.netloc.partition(":")
        host = host or "localhost"
        port = int(port) or 5432

        async with SITL(host, port) as sitl:
            async with create_daemon_task_group() as task_group:
                task_group.start_soon(self._worker, sitl)
                yield self

    def __init__(self, preset: str = "default"):
        """Constructor.

        Parameters:
            preset: name of a preset from the PRESETS attribute of the class
                that determines how often the driver should poll the downlink
                with null packets and how it should handle packet resending
        """
        try:
            self.apply_preset(preset)
        except KeyError:
            self.apply_preset("default")

        # TODO(ntamas): what if the in_queue is full?
        self._in_queue_tx, self._in_queue_rx = create_memory_object_stream[CRTPPacket](
            256
        )
        self._out_queue_tx, self._out_queue_rx = create_memory_object_stream[
            CRTPPacket
        ](1)

    def apply_preset(self, name: str) -> None:
        """Applies a preset strategy to the given connection to control how
        often should the driver pull the downlink with null packets.

        This method can be called with an active connection; the new preset
        will take effect as soon as possible.
        """
        try:
            preset = self.PRESETS[name]
        except KeyError:
            raise KeyError("no such preset: {0}".format(name)) from None

        self.polling_strategy = preset[0]()

    @property
    def is_safe(self) -> bool:
        return True

    @property
    def link_quality(self) -> ObservableValue[float]:
        global _link_quality
        if _link_quality is None:
            _link_quality = ObservableValue.constant(1.0)
        return _link_quality

    @property
    def name(self) -> str:
        return "SITL"

    async def receive_packet(self) -> CRTPPacket:
        """Receives a single CRTP packet.

        Returns:
            the next CRTP packet that was received
        """
        return await self._in_queue_rx.receive()

    async def send_packet(self, packet: CRTPPacket) -> None:
        """Sends a CRTP packet.

        Parameters:
            packet: the packet to send
        """
        await self._out_queue_tx.send(packet)

    async def _worker(self, sitl: SITL) -> None:
        """Worker task that runs continuously and handles the sending and
        receiving of packets to/from a given SITL instance.

        Parameters:
            sitl: the SITL instance to use
        """
        null_packet = outbound_packet = CRTPPacket.null()
        delay_before_next_null_packet = 0.01

        while True:
            to_send = outbound_packet.to_bytes()
            await sitl.send_bytes(to_send)

            response = None
            with move_on_after(0.02):
                response = await sitl.receive_bytes()

            if response is not None:
                inbound_packet = CRTPPacket.from_bytes(response)
                await self._in_queue_tx.send(inbound_packet)
            else:
                response = b"\xff"

            # Figure out how much to wait before the next null packet is sent
            delay_before_next_null_packet = self.polling_strategy(response, to_send)
            if delay_before_next_null_packet > 0:
                # Wait for a given number of seconds
                outbound_packet = null_packet
                with move_on_after(delay_before_next_null_packet):
                    outbound_packet = await self._out_queue_rx.receive()
            elif delay_before_next_null_packet < 0:
                # Wait indefinitely
                outbound_packet = await self._out_queue_rx.receive()
            else:
                # Poll the outbound queue; send a null packet if the queue is
                # empty
                try:
                    outbound_packet = self._out_queue_rx.receive_nowait()
                except WouldBlock:
                    outbound_packet = null_packet



from typing import Callable, List, Union

__all__ = (
    "BackoffPollingStrategy",
    "DefaultPollingStrategy",
    "DefaultResendingStrategy",
    "NoPollingStrategy",
    "PatientResendingStrategy",
    "PollingStrategy",
    "ResendingStrategy",
    "ResendingStrategyResult",
)


#: Type specification for polling strategies
PollingStrategy = Callable[[bytes, bytes], float]

#: Type specification for result objects of a resending strategy
ResendingStrategyResult = Union[str, float]

#: Type specification for resending strategies
ResendingStrategy = Callable[[bool, bytes], ResendingStrategyResult]


class DefaultPollingStrategy:
    """Default polling strategy for Crazyflie connections.

    The polling strategy kicks in when we have no packets to send to the
    Crazyflie. Since the downstream link has to be pulled explicitly, we need
    to decide how much time to wait until we send a null packet to the
    Crazyflie.

    This function implements the default behaviour, which is as follows: send
    the next null packet immediately if we have received non-empty packets
    recently from the drone, otherwise poll the link at 100 Hz. We switch back
    to polling at 100 Hz if the last ten packets that we have received from
    the drone were empty.

    This polling strategy is consistent with the default behaviour of the
    Crazyflie radio protocol as implemented in the official Python API.
    """

    def __init__(self, frequency: float = 100):
        """Constructor.

        Parameters:
            frequency: the polling frequency if the link is idle
        """
        self._empty_packet_counter = 0
        self._is_idle = False
        self._interval = 1 / frequency

    def __call__(self, data: bytes, sent: bytes) -> float:
        """Returns the time to wait before we send the next null packet.

        Parameters:
            data: the last packet that we have received from the drone
            sent: the last packet that we have sent to the drone

        Returns:
            the proposed time to wait before sending the next null packet,
            in seconds; negative numbers mean to wait indefinitely
        """
        # The packet starts with the header so its length is always at least
        # 1. The packet is effectively empty if its length is less than 2.
        if len(sent) > 1 or len(data) > 1:
            self._empty_packet_counter = 0
            self._is_idle = False
        elif not self._is_idle:
            self._empty_packet_counter += 1
            if self._empty_packet_counter >= 10:
                self._is_idle = True
        return self._interval if self._is_idle else 0


class NoPollingStrategy:
    """Dummy polling strategy that never sends null packets to pull the
    downlink; it is assumed that we expect responses to our own packets only.
    """

    def __call__(self, received: bytes, sent: bytes) -> float:
        """Returns the time to wait before we send the next null packet.

        Parameters:
            data: the last packet that we have received from the drone
            sent: the last packet that we have sent to the drone

        Returns:
            -1 because it means "wait indefinitely" for the caller.
        """
        return -1


class BackoffPollingStrategy:
    """Lenient polling strategy for Crazyflie connections that gradually
    increases the time spent between consecutive null packets if there is no
    traffic to avoid saturating the link if there are many drones to communicate
    with.

    The strategy works as follows. If the last packet that we have received from
    the drone was non-empty, we poll the next packet immediately in case there
    is a continuation. Otherwise, we start polling at 100 Hz and double the
    delay after each polling attempt that yielded no response from the drone. If
    the delay reaches 250 msec, we do not increase it any further.
    """

    def __init__(self, initial_delay: float = 0.01, max_delay: float = 0.25):
        """Constructor.

        Parameters:
            initial_delay: the initial delay between consecutive polls, in
                seconds
            max_delay: maximum delay between consecutive polls.
        """
        self._empty_packet_counter = 0
        self._is_idle = False

        self._initial_delay = float(initial_delay)
        self._max_delay = float(max_delay)
        self._delay = self._initial_delay

    def __call__(self, data: bytes, sent: bytes) -> float:
        """Returns the time to wait before we send the next null packet.

        Parameters:
            data: the last packet that we have received from the drone
            sent: the last packet that we have sent to the drone

        Returns:
            the proposed time to wait before sending the next null packet, in
            seconds; negative numbers mean to wait indefinitely
        """
        # The packet starts with the header so its length is always at least
        # 1. The packet is effectively empty if its length is less than 2.
        if len(sent) > 1 or len(data) > 1:
            self._empty_packet_counter = 0
            self._is_idle = False
        elif not self._is_idle:
            self._empty_packet_counter += 1
            if self._empty_packet_counter >= 10:
                self._is_idle = True
                self._delay = self._initial_delay

        if self._is_idle:
            result = self._delay
            if self._delay < self._max_delay:
                self._delay = min(self._delay * 2, self._max_delay)
            return result
        else:
            return 0


class DefaultResendingStrategy:
    """Default packet resending strategy for Crazyflie connections.

    Radio packets to the Crazyflie occasionally fail to go through. This may be
    due to interference or due to the fact that the Crazyflie is not in the
    radio range any more. To distinguish between the two, this strategy resends
    failed packets and counts the number of consecutive failures; a link error
    is reported if the number of consecutive failures reaches a configurable
    threshold (100 attempts by default).
    """

    def __init__(self, attempts: int = 100):
        """Constructor.

        Parameters:
            attempts: number of sending attempts before giving up
        """
        self._remaining = self._attempts = 100

    def __call__(self, ack: bool, data: bytes) -> ResendingStrategyResult:
        """Decides whether the packet should be resent.

        Parameters:
            ack: whether the packet was acknowledged by the radio
            data: the packet that was sent

        Returns:
            whether the packet should be accepted (`accept`), re-sent again
            after a delay (returns the desired delay in seconds in this case),
            or the connection should be dropped (`stop`)
        """
        if ack:
            self._remaining = self._attempts
            return "accept"
        else:
            self._remaining -= 1
            return 0 if self._remaining > 0 else "stop"


class PatientResendingStrategy:
    """More patient packet resending strategy for Crazyflie connections."""

    _remaining: int
    _attempts: int
    _num_errors: int
    _scheduled: List[float]

    def __init__(self, attempts: int = 50):
        """Constructor.

        Parameters:
            attempts: number of sending attempts before giving up
        """
        self._remaining = self._attempts = 50
        self._num_errors = 0
        self._schedule = [0, 0, 0, 0, 0, 0, 0.01, 0.01, 0.01, 0.01, 0.01, 0.02]

    def __call__(self, ack: bool, data: bytes) -> ResendingStrategyResult:
        """Decides whether the packet should be resent.

        Parameters:
            ack: whether the packet was acknowledged by the radio
            data: the packet that was sent

        Returns:
            whether the packet should be accepted (`accept`), re-sent again
            after a delay (returns the desired delay in seconds in this case),
            or the connection should be dropped (`stop`)
        """
        if ack:
            self._remaining = self._attempts
            self._num_errors = 0
            return "accept"
        else:
            self._remaining -= 1
            self._num_errors += 1
            if self._remaining > 0:
                if len(self._schedule) > self._num_errors:
                    return self._schedule[self._num_errors]
                else:
                    return self._schedule[-1]
            else:
                return "stop"



from contextlib import asynccontextmanager
from typing import AsyncIterator, Iterable, Optional, Union

from aiocflib.drivers.crazyradio import RadioConfiguration
from aiocflib.crtp.crtpstack import CRTPPacket, CRTPPortLike
from aiocflib.crtp.device import _handle_data_argument
from aiocflib.crtp.drivers.radio import SharedCrazyradio
from aiocflib.crtp.exceptions import WrongURIType
from aiocflib.utils.addressing import DEFAULT_RADIO_BROADCAST_ADDRESS, parse_radio_uri


class _Broadcaster:
    """Class that provides functions for broadcasting CRTP packets to all
    devices listening at a specific broadcast address.
    """

    def __init__(self, uri: str, radio, configuration: RadioConfiguration):
        """Constructor.

        Creates a broadcaster object that uses the given shared radio object and
        radio configuration.

        Do not use directly; use the Broadcaster_ context manager instead.
        """
        self._uri = uri
        self._radio = radio
        self._configuration = configuration

    @property
    def uri(self) -> str:
        """URI of the target of the broadcaster."""
        return self._uri

    async def send_packet(
        self,
        *,
        port: CRTPPortLike,
        channel: int = 0,
        data: Optional[Union[int, bytes, Iterable[Union[int, bytes]]]] = None,
    ) -> None:
        """Broadcasts a packet with the given CRTP port, channel and the given
        body.

        Parameters:
            port: the CRTP port to send the packet to
            channel: the CRTP channel to send the packet to
            data: the body of the request packet
        """
        packet = CRTPPacket(port=port, channel=channel)
        packet.data = _handle_data_argument(data)
        await self.send_bytes(packet.to_bytes())

    async def send_bytes(self, data: bytes) -> None:
        """Broadcasts some raw bytes to the associated broadcast address.

        Parameters:
            data: the data to broadcast
        """
        await self._radio.configure_send_and_receive_bytes(self._configuration, data)


@asynccontextmanager
async def Broadcaster(uri: str) -> AsyncIterator[_Broadcaster]:
    """Async context manager that creates an object that allows the user to
    send broadcast packets to the given radio URI.

    Parameters:
        uri: the URI describing the data link where the broadcast packets
            should be sent to. It must be a radio URL with the radio index,
            the channel and the data rate specified; the address may be
            omitted and defaults to the Crazyradio broadcast address
    """
    try:
        parts = parse_radio_uri(uri, default_address=DEFAULT_RADIO_BROADCAST_ADDRESS)
    except ValueError:
        raise WrongURIType from None

    index = parts.pop("index")
    configuration = RadioConfiguration(**parts)

    async with SharedCrazyradio(index) as radio:
        yield _Broadcaster(configuration.to_uri(index), radio, configuration)


async def test():
    from aiocflib.bootloader.target import BootloaderTargetType
    from aiocflib.bootloader.types import BootloaderCommand
    from aiocflib.crtp.crtpstack import CRTPPort, LinkControlChannel

    async with Broadcaster("radio://0/80/2M") as broadcaster:
        print(repr(broadcaster.uri))
        await broadcaster.send_packet(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            data=(BootloaderTargetType.NRF51, BootloaderCommand.SHUTDOWN),
        )


if __name__ == "__main__":
    from anyio import run

    run(test)



from typing import Callable, Generic, Optional, TypeVar, Union, overload


T = TypeVar("T")


class Registry(Generic[T]):
    """Generic registry object that associates string keys to factory
    functions.
    """

    def __init__(self):
        """Constructor."""
        self._items = {}

    def find(self, key: str, **kwds) -> T:
        """Finds an item in the registry with the given key.

        Parameters:
            key: the key to look up

        Keyword arguments:
            default: the default value to return if there is no value
                associated to the key

        Returns:
            the value associated to the key, or the default value if there is
            no such key and a default value is provided

        Raises:
            KeyError: if there is no such item for the given key, and no default
                value was provided
        """
        try:
            return self._items[key]
        except KeyError:
            if "default" in kwds:
                return kwds["default"]
            else:
                raise

    @overload
    def register(self, key: str) -> Callable[[T], T]:
        ...

    @overload
    def register(self, key: str, value: T) -> T:
        ...

    def register(
        self, key: str, value: Optional[T] = None
    ) -> Union[T, Callable[[T], T]]:
        """When called with two arguments, associates an item to the given key
        and checks for duplicates to ensure that already registered items cannot
        be overridden. When called with a single argument, returns a decorator
        that can be applied to a value to register it.

        Parameters:
            key: the key to register the item to
            value: the value to register, or `None` to return a decorator
        """
        if value is not None:
            existing = self._items.get(key)
            if existing:
                raise ValueError(
                    "Name {0!r} is already registered for {1!r}".format(key, existing)
                )
            self._items[key] = value
            return value
        else:

            def decorator(item):
                if item is not None:
                    return self.register(key, item)
                else:
                    raise ValueError("None cannot be registered")

            return decorator
